#!python
from __future__ import division,print_function
from pyPRISM.omega.Omega import Omega
import numpy as np
from math import exp,sin,cos,sqrt
from scipy.optimize import root

class DiscreteKoyama(Omega):
    r'''Semi-flexible Koyama-based intra-molecular correlation function

    **Mathematial Definition**

    .. math::
    
        \hat{\omega}(k) = \frac{\sin(Bk)}{Bk}\exp(-A^2k^2)

    .. math::
         
         A^2 = \frac{\langle r_{\alpha,\beta}^2 \rangle (1-C)}{6} 
	
    .. math::
         
         B^2 =  C \langle r_{\alpha,\beta}^2 \rangle  
    
    .. math::
         
         C^2 =  \frac{1}{2}\left(5-3\frac{ \langle r_{\alpha,\beta}^4 \rangle}{ \langle r_{\alpha,\beta}^2 \rangle}\right)


    **Variable Definitions**
        
        - :math:`\hat{\omega}(k)` 
            *intra*-molecular correlation function at wavenumber :math:`k`

        - :math:`\langle r_{\alpha,\beta}^2 \rangle`
            second moment of the distance distribution between sites
            :math:`\alpha` and :math:`\beta`. Please see equation (17) of the
            Reference [1] cited below for the mathematical representation.       
 
        - :math:`\langle r_{\alpha,\beta}^4 \rangle`
            fourth moment of the distance distribution between sites
            :math:`\alpha` and :math:`\beta`. Please see equations (18-24) of
            the reference cited below for the mathematical representation.       


    **Description**
        
        The discrete Koyama :math:`\hat{\omega}(k)` was developed to
        represent a wormlike chain with semiflexibility. This scheme
        interpolates between the rigid-rod and the Gaussian chain limits
        to represent a chain with a given persistence length. This 
        form for :math:`\hat{\omega}(k)` has been shown to match the
        structure of molecular dynamics simulations of Kremer-Grest 
        style bead-spring polymer models. 
    
    References
    ----------
    #. Honnell, K.G., J.G. Curro, and K.S. Schweizer, LOCAL-STRUCTURE OF
       SEMIFLEXIBLE POLYMER MELTS. Macromolecules, 1990. 23(14): p. 3496-3505.
       [`link <https://doi.org/10.1021/ma00216a018>`__]


    Example
    -------
    .. code-block:: python

        import pyPRISM
        import numpy as np
        import matplotlib.pyplot as plt

        #calculate Fourier space domain and omega values
        domain = pyPRISM.domain(dr=0.1,length=1000)
        omega  = pyPRISM.omega.DiscreteKoyama(sigma=1.0,l=1.0,length=100,lp=1.43)
        x = domain.k
        y = omega.calculate(x)
        
        #plot the results using matplotlib
        plt.plot(x,y)
        plt.gca().set_xscale("log", nonposx='clip')
        plt.gca().set_yscale("log", nonposy='clip')

        plt.show()
        
        #define a PRISM system and set omega(k) for type A
        sys = pyPRISM.System(['A','B'],kT=1.0)
        sys.domain = pyPRISM.Domain(dr=0.1,length=1024)
        sys.omega['A','A']  = pyPRISM.omega.DiscreteKoyama(sigma=1.0,l=1.0,length=100,lp=1.43)
    
    '''
    def __init__(self,sigma,l,length,lp):
        r'''Constructor

        Arguments
        ---------
        sigma: float
            contact distance between sites (i.e. site diameter)

        l: float
            bond length
            
        length: float
            number of monomers/sites in the chain

        lp: float
            persistence length of chain
        '''
        self.sigma   = sigma
        self.length  = int(length)
        self.l       = l
        self.lp      = lp
        self.cos0    = 1 - sigma*sigma/(2.0 * l * l)
        self.value   = None

        if self.lp<4.0/3.0:
            raise ValueError('DiscreteKoyama does not support persistence lengths < 4.0/3.0.')
        elif self.lp == 4.0/3.0:
            self.epsilon = 0.0
            self.cos1 = 0.5*(self.cos0-1.0)/(self.cos0 + 1.0)
            self.cos2 = (self.cos0**(3.0) + 1)/(3*self.cos0 + 3)
        else:

            self.cos1 = l/lp - 1
            funk = lambda e: self.cos_avg(e) - self.cos1
            result  = root(funk,1.0)

            if result.success != True:
                raise ValueError('DiscreteKoyama initialization failure. Could not solve for bending energy.')

            self.epsilon = result.x
            self.cos2 = self.cos_sq_avg(self.epsilon)
    def cos_avg(self,epsilon):
        '''First moment of bond angle distribution'''
        e = epsilon
        cos0 = self.cos0
        return 1/e  - ( exp(e) + cos0*exp(-e*cos0) )/( exp(e) - exp(-e*cos0) )
    
    def cos_sq_avg(self,epsilon):
        '''Second moment of bond angle distribution'''
        e = epsilon
        cos0 = self.cos0
        cos1 = self.cos_avg(epsilon)
        return (2/e)*cos1 + ( exp(e) - cos0*cos0*exp(-e*cos0) )/( exp(e) - exp(-e*cos0) )
    
    def kernel_base(self,n):
        ''' Calculates the second and fourth moments of the site separate distance distributions

        .. note::

            See Equation 18 in Reference [1] for more details.

        Arguments
        ---------
        n: int
            Integer separation distance along chain.

        '''
        l = self.l
        q = -self.cos1
        p = (3*self.cos2 - 1)/2
        
        D  = n * n * ((1 + q)/(1 - q))**(2.0) 
        D -= n*(1 + (2*q/(1-q)**(3.0)) * (6 + 5*q + 3*q*q) - 4*p/(1-p)*((1 + q)/(1 - q))**(2.0))
        D += 2*q/(1-q)**(4.0) * (4 + 11*q + 12*q*q)
        D -= 4*p/(1-p) * (1 + 8*q/(1-q)**(3.0) + p/(1-p)*((1 + q)/(1 - q))**(2.0))
        D -= q**(n) * 8*q/(1-q)**(3.0) * (n*(1 + 3*q))
        D -= q**(n) * 8*q/(1-q)**(3.0) * ((1 + 2*q + 3*q*q)/(1-q))
        D -= q**(n) * 8*q/(1-q)**(3.0) * (-2*p/(q-p)**(2.0) *(n*(1-q)*(q-p)+2*q*q-q*p-p))
        D -= 6*q**(2*n+2)/(1-q)**(4.0)
        D += p**(n) * (4/(1-p) * (1 + 8*q/(1-q)**(3.0) - ((1+q)/(1-q))**2.0 * (1 - p/(1-p)) ))
        D -= p**(n) * (16*q*q/(1-q)**(3.0) * (1/(q-p)**(2.0))*(q+q*q-2*p))
        D *= 2/3
        
        r2 = n*l*l*((1-self.cos1)/(1+self.cos1) + 2*self.cos1/n * (1-(-self.cos1)**(n))/(1 + self.cos1)**(2.0))
        r4 = r2*r2 + l*l*l*l*D

        return r2,r4
        

    def koyama_kernel_fourier(self,k,n):
        '''Kernel for calculating omega in Fourier-Space

        .. note::

            See Equation 16 in Reference [1] for more details.

        Arguments
        ---------
        k: np.ndarray, float
            array of wavenumber values to calculate :math:`\omega` at

        n: int
            Integer separation distance along chain.
        '''
        r2,r4 = self.kernel_base(n)
        try:
            C = sqrt(0.5 * (5 - 3*r4/(r2*r2)))
            B = sqrt(C*r2)
            Asq = r2*(1-C)/6 #taking the square root results in many domain errors
        except ValueError as e:
            raise ValueError('Bad chain parameters. (Try reducing epsilon)')
            
        return np.sin(B*k)/(B*k) * np.exp(-Asq*k*k)

    def koyama_kernel_real(self,r,n):
        '''Kernel for calculating omega in Real-Space

        .. note::

            See Equation 12 in Reference [1] for more details.

        Arguments
        ---------
        r: np.ndarray, float
            array of real-space positions to calculate :math:`\omega` at

        n: int
            Integer separation distance along chain.
        '''
        r2,r4 = self.kernel_base(n)
        
        try:
            C = sqrt(0.5 * (5 - 3*r4/(r2*r2)))
            B = sqrt(C*r2)
            Asq = r2*(1-C)/6 #taking the square root results in many domain errors
        except ValueError as e:
            raise ValueError('Bad chain parameters. (Try reducing epsilon)')

        omega_ag = (1.0/(8*np.pi**(3.0/2.0)*np.sqrt(Asq)*B*r))*(np.exp(-(r-B)**2.0/(4.0*Asq))-np.exp(-(r+B)**2.0/(4.0*Asq)))
        
        return omega_ag

    def density_correction_kernel(self,r):
        '''Correction for density due to non-physical overlaps

        .. note::

            See Equation 28 in Reference [1] for more details.

        Arguments
        ---------
        r: np.ndarray, float
            array of real-space positions to calculate :math:`\omega` at
        '''
        
        factor1 = np.pi*self.sigma**(3.0)*(1-3.0*r/(2.0*self.sigma)+r**(3.0)/(2.0*self.sigma**3.0))/6.0
        factor2 = np.zeros_like(r) 
        for i in range(1,self.length-1):
            for j in range(i+2,self.length+1):
                n = abs(i - j)
                factor2 += self.koyama_kernel_real(r=r,n=n)
        factor3 = 4.0*np.pi*r**2.0

        return factor1*factor2*factor3

    def density_correction(self,npts=1000):
        '''Correction for density due to non-physical overlaps

        .. note::

            See Equation 28 in Reference [1] for more details.

        Arguments
        ---------
        npts: int
            number of points to use in numerical integral 
        '''
        
        r = np.linspace(0.0001,self.sigma,npts)
        integral = np.trapz(self.density_correction_kernel(r),r)
        delta_N = integral/(self.length*np.pi*self.sigma**3.0/6.0)

        return delta_N

    def __repr__(self):
        return '<Omega: Koyama>'
    
    def calculate(self,k):
        '''Return value of :math:`\hat{\omega}` at supplied :math:`k`

        Arguments
        ---------
        k: np.ndarray, float
            array of wavenumber values to calculate :math:`\omega` at
        
        '''
        self.value = np.zeros_like(k)
        
        for i in range(1,self.length-1):
            for j in range(i+1,self.length):
                n = abs(i - j)
                self.value += self.koyama_kernel_fourier(k=k,n=n)
        self.value *= 2/self.length
        self.value += 1.0
        
        return self.value





#!python
from __future__ import division,print_function
from pyPRISM.core.PairTable import PairTable
from pyPRISM.core.MatrixArray import MatrixArray
from pyPRISM.core.Space import Space
from pyPRISM.calculate.structure_factor import structure_factor
import numpy as np

def solvation_potential(PRISM,closure='HNC'):
    r'''Calculate the pairwise decomposed medium-induced solvation potential
    
        
    Parameters
    ----------
    PRISM: pyPRISM.core.PRISM
        A **solved** PRISM object.

    closure: str ('PY' or 'HNC')
        closure used to derive the potential 

    
    Returns
    -------
    psi: pyPRISM.core.MatrixArray
        MatrixArray of the *Real-space* solvation potentials


    **Mathematical Definition**

    .. math::

        \text{PY: } \Delta \hat{\Psi}^{PY}(k) = - k_B T \ln(1 + \hat{C}(k)\hat{S}(k)\hat{C}(k))

    .. math::

        \text{HNC: } \Delta \hat{\Psi}^{HNC}(k) = - k_B T \hat{C}(k)\hat{S}(k)\hat{C}(k)

    
    **Variable Definitions**

        - :math:`\Delta \hat{\Psi}^{PY}`, :math:`\Delta \hat{\Psi}^{HNC}`
            Percus-Yevick and Hypernetted Chain derived pairwise decomposed
            solvation potentials, each described as a  :class:`.MatrixArray`.
            This implies that the multiplication in the above equation is
            actually *matrix* multiplication and the individual solvation
            potentials are extracted as curves of the MatrixArrays. Note that
            the solvation potential MatrixArrays are inverted back to
            Real-space for use. 

        - :math:`\hat{C}(k)`
            Direct correlation function :class:`.MatrixArray` at a wavenumber :math:`k`

        - :math:`\hat{S}(k)`
            Structure factor :class:`.MatrixArray` at a wavenumber :math:`k`

        - :math:`k_B T`
            Thermal temperature written as the product of the Boltzmann
            constant and temperature.

    **Description**

        The solvation potential (:math:`\Delta \hat{\Psi}`) mathematically
        describes how a given surrounding medium perturbs the site-site pairwise
        interactions of a molecule.

        This calculation is the foundation of the Self-Consistent PRISM
        formalism. See :ref:`SCPRISM` for more information.

    .. warning::

        Passing an unsolved PRISM object to this function will still produce
        output based on the default values of the attributes of the PRISM
        object.
    
    References
    ----------

    #. Grayce, Schweizer, Solvation potentials for macromolecules, J. Chem.
       Phys., 1994 100 (9) 6846 [`link <https://doi.org/10.1063/1.467044>`__]
    
    #. Schweizer, Honnell, Curro, Reference interaction site model theory of
       polymeric liquids: Self-consistent formulation and nonideality effects in
       dense solutions and melts, J. Chem. Phys., 1992 96 (4) 3211 [`link
       <https://doi.org/10.1063/1.461965>`__]

    Example
    -------
    .. code-block:: python

        import pyPRISM

        sys = pyPRISM.System(['A','B'])

        # ** populate system variables **
        
        PRISM = sys.createPRISM()

        PRISM.solve()

        psi = pyPRISM.calculate.solvation_potential(PRISM)

        psi_BB = psi['B','B']
    
    '''
    
    assert PRISM.sys.rank>1,'the psi calculation is only valid for multicomponent systems'
    
    if PRISM.directCorr.space == Space.Real:
        PRISM.domain.MatrixArray_to_fourier(PRISM.directCorr)

    if PRISM.totalCorr.space == Space.Real:
        PRISM.sys.domain.MatrixArray_to_fourier(PRISM.totalCorr)
        
    if PRISM.omega.space == Space.Real:
        PRISM.sys.domain.MatrixArray_to_fourier(PRISM.omega)

    structureFactor = structure_factor(PRISM)
    #(PRISM.totalCorr*PRISM.sys.pairDensityMatrix + PRISM.omega)/PRISM.sys.siteDensityMatrix

    if closure == 'HNC':
        psi = PRISM.directCorr.dot(structureFactor).dot(PRISM.directCorr) * -PRISM.sys.kT 
    elif closure == 'PY':
        psi = PRISM.directCorr.dot(structureFactor).dot(PRISM.directCorr)
        psi.data = np.log(1 + psi.data) * -PRISM.sys.kT

    PRISM.sys.domain.MatrixArray_to_real(psi)

    return psi



#!python
from pyPRISM.core.MatrixArray import MatrixArray
import numpy as np
import unittest

class MatrixArray_TestCase(unittest.TestCase): 
    def test_assign(self):
        '''Can we create and assign values?'''
        length = 100
        rank = 8
        MA = MatrixArray(length=length,rank=rank)
        
        # Make sure the array starts as all zeros
        array = np.zeros((length,rank,rank))
        np.testing.assert_array_almost_equal(MA.data,array)
        
        #  Test assignment (especially off diagonal)
        array = np.zeros((length,rank,rank))
        MA['B','B'] = np.ones(length)
        MA['B','C'] = np.ones(length)*3.0
        array[:,1,1] = np.ones(length)
        array[:,1,2] = np.ones(length)*3.0
        array[:,2,1] = np.ones(length)*3.0
        np.testing.assert_array_almost_equal(MA.data,array)
        
    def test_div(self):
        '''Can we truediv and itruediv?'''
        
        length = 100
        rank = 3
        (MA1,MA2),(array1,array2) = self.set_up_test_arrays(length,rank)
        
        MA2['A','C'] = np.ones(length)*2.0
        MA2['A','B'] = np.ones(length)*3.0
        MA2['B','C'] = np.ones(length)*5.0
        MA2['A','A'] += 1.0
        
        array2[:,0,2] = array2[:,2,0] = np.ones(length)*2.0
        array2[:,0,1] = array2[:,1,0] = np.ones(length)*3.0
        array2[:,1,2] = array2[:,2,1] = np.ones(length)*5.0
        array2[:,0,0] += 1.0
        
        ## Test Add
        MA3 = MA1 / MA2
        array3 = array1 / array2
        MA3 = MA3 / 542.345
        array3 = array3 / 542.345
        np.testing.assert_array_almost_equal(MA1.data,array1)
        np.testing.assert_array_almost_equal(MA2.data,array2)
        np.testing.assert_array_almost_equal(MA3.data,array3)
        
        ## Test iAdd
        MA3 /= MA2
        array3 /= array2
        MA3 /= 324
        array3 /= 324
        np.testing.assert_array_almost_equal(MA1.data,array1)
        np.testing.assert_array_almost_equal(MA2.data,array2)
        np.testing.assert_array_almost_equal(MA3.data,array3)
        
    def set_up_test_arrays(self,length=100,rank=3):
        ''' Helper for set up arrays to test math'''
        MA1 = MatrixArray(length=length,rank=rank)
        MA1['A','A'] = np.ones(length)*2.0
        MA1['B','B'] = np.ones(length)
        MA1['C','C'] = np.ones(length)*3.0
        MA1['B','C'] = np.ones(length)*3.0
        
        MA2 = MatrixArray(length=length,rank=rank)
        MA2['A','A'] = np.arange(length)
        MA2['B','B'] = np.ones(length)*2.0
        MA2['C','C'] = np.ones(length)*-3.0
        
        array1 = np.zeros((length,rank,rank))
        array1[:,0,0] = np.ones(length)*2.0
        array1[:,1,1] = np.ones(length)
        array1[:,2,2] = np.ones(length)*3.0
        array1[:,1,2] = np.ones(length)*3.0
        array1[:,2,1] = np.ones(length)*3.0
        
        array2 = np.zeros((length,rank,rank))
        array2[:,0,0] = np.arange(length)
        array2[:,1,1] = np.ones(length)*2.0
        array2[:,2,2] = np.ones(length)*-3.0
        return (MA1,MA2),(array1,array2)
        
    def test_add(self):
        '''Can we add and iadd?'''
        
        length = 100
        rank = 3
        (MA1,MA2),(array1,array2) = self.set_up_test_arrays(length,rank)
        
        ## Test Add
        MA3 = MA1 + MA2
        array3 = array1 + array2
        MA3 = MA3 + 2.53
        array3 = array3 + 2.53
        np.testing.assert_array_almost_equal(MA1.data,array1)
        np.testing.assert_array_almost_equal(MA2.data,array2)
        np.testing.assert_array_almost_equal(MA3.data,array3)
        
        ## Test iAdd
        MA3 += MA2
        MA3 += 435.43
        array3 += array2
        array3 += 435.43
        np.testing.assert_array_almost_equal(MA1.data,array1)
        np.testing.assert_array_almost_equal(MA2.data,array2)
        np.testing.assert_array_almost_equal(MA3.data,array3)
        
    def test_sub(self):
        '''Can we sub and isub?'''
        
        length = 100
        rank = 3
        (MA1,MA2),(array1,array2) = self.set_up_test_arrays(length,rank)
        
        ## Test Add
        MA3 = MA1 - MA2
        array3 = array1 - array2
        MA3 = MA3 - 852.32
        array3 = array3 - 852.32
        np.testing.assert_array_almost_equal(MA1.data,array1)
        np.testing.assert_array_almost_equal(MA2.data,array2)
        np.testing.assert_array_almost_equal(MA3.data,array3)
        
        ## Test iAdd
        MA3 -= MA2
        array3 -= array2
        MA3 -= 2
        array3 -= 2
        np.testing.assert_array_almost_equal(MA1.data,array1)
        np.testing.assert_array_almost_equal(MA2.data,array2)
        np.testing.assert_array_almost_equal(MA3.data,array3)
        
    def test_mul(self):
        '''Can we mul and imul?'''
        
        length = 100
        rank = 3
        (MA1,MA2),(array1,array2) = self.set_up_test_arrays(length,rank)
        
        ## Test Add
        MA3 = MA1 * MA2
        array3 = array1 * array2
        MA3 = MA3 * 542.345
        array3 = array3 * 542.345
        np.testing.assert_array_almost_equal(MA1.data,array1)
        np.testing.assert_array_almost_equal(MA2.data,array2)
        np.testing.assert_array_almost_equal(MA3.data,array3)
        
        ## Test iAdd
        MA3 *= MA2
        array3 *= array2
        MA3 *= 324
        array3 *= 324
        np.testing.assert_array_almost_equal(MA1.data,array1)
        np.testing.assert_array_almost_equal(MA2.data,array2)
        np.testing.assert_array_almost_equal(MA3.data,array3)
        
    def test_invert(self):
        '''Can we matrix invert?'''
        
        length = 100
        rank = 3
        (MA1,_),(array1,_) = self.set_up_test_arrays(length,rank)
        
        MA2 = MA1.invert(inplace=False)
        
        array2 = np.empty_like(array1)
        for i in range(length):
            array2[i] = np.linalg.inv(array1[i])
        
        np.testing.assert_array_almost_equal(MA1.data,array1)
        np.testing.assert_array_almost_equal(MA2.data,array2)
        
        MA1.invert(inplace=True)
        np.testing.assert_array_almost_equal(MA1.data,MA2.data)
        
    def test_dot(self):
        '''Can we matrix multiply?'''
        
        length = 100
        rank = 3
        (MA1,MA2),(array1,array2) = self.set_up_test_arrays(length,rank)
        
        MA3 = MA1.dot(MA2,inplace=False)
        
        array3 = np.empty_like(array1)
        for i in range(length):
            array3[i] = np.dot(array1[i],array2[i])
        
        np.testing.assert_array_almost_equal(MA1.data,array1)
        np.testing.assert_array_almost_equal(MA2.data,array2)
        np.testing.assert_array_almost_equal(MA3.data,array3)
        
        MA1.dot(MA2,inplace=True)
        np.testing.assert_array_almost_equal(MA1.data,MA3.data)
        
    def test_itercurve(self):
        ''' Can we iterate over the curves?'''
        length = 100
        rank = 5
        (MA1,_),(array1,_) = self.set_up_test_arrays(length,rank)
        
        ncols = 0
        for (i,j),(t1,t2),col in MA1.itercurve():
            # with self.subTest(i=i,j=j):
            #     np.testing.assert_array_almost_equal(col,array1[:,i,j])
            np.testing.assert_array_almost_equal(col,array1[:,i,j])
            ncols+=1
        self.assertEqual(ncols,rank*(rank+1)//2)
                
    def test_itercurve_assign(self):
        ''' Can we assign as we iterate over the curves?'''
        length = 100
        rank = 4
        (MA1,_),(array1,_) = self.set_up_test_arrays(length,rank)
        
        ncols = 0
        for (i,j),(t1,t2),col in MA1.itercurve():
            MA1[t1,t2] = np.ones(length)*i + j/2.0
            array1[:,i,j] = np.ones(length)*i + j/2.0
            array1[:,j,i] = np.ones(length)*i + j/2.0
            ncols += 1 
            
        np.testing.assert_array_almost_equal(MA1.data,array1)
        self.assertEqual(ncols,rank*(rank+1)//2)
            
    def test_get(self):
        ''' Can we access the data indices?'''
        length = 100
        rank = 3
        types = ['A','B','C']
        (MA1,_),(array1,_) = self.set_up_test_arrays(length,rank)
        
        MA1.types = types
        
        np.testing.assert_array_almost_equal(MA1['A','A'],MA1.get(0,0))
        np.testing.assert_array_almost_equal(MA1['A','B'],MA1.get(0,1))
        np.testing.assert_array_almost_equal(MA1['A','C'],MA1.get(0,2))
        np.testing.assert_array_almost_equal(MA1['B','A'],MA1.get(1,0))
        np.testing.assert_array_almost_equal(MA1['B','B'],MA1.get(1,1))
        np.testing.assert_array_almost_equal(MA1['B','C'],MA1.get(1,2))
        np.testing.assert_array_almost_equal(MA1['C','A'],MA1.get(2,0))
        np.testing.assert_array_almost_equal(MA1['C','B'],MA1.get(2,1))
        np.testing.assert_array_almost_equal(MA1['C','C'],MA1.get(2,2))
        
            
if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(MatrixArray_TestCase)
    unittest.TextTestRunner(verbosity=2).run(suite)



from pyPRISM.core.Table import Table
from pyPRISM.core.MatrixArray import MatrixArray
from pyPRISM.core.Space import Space
from itertools import product
import numpy as np
import copy

class PairTable(Table):
    '''Container for data that is keyed by pairs of types

    **Description**

        Since PRISM is a theory based in *pair*-correlation functions, it
        follows that many of the necessary parameters of the theory are
        specified between the pairs of types. This goal of this container is to
        make setting, getting, and checking these data easy.
        
        Setter/getter methods have been set up to set groups of types
        simultaneously. This allows for the rapid construction of datasets
        where many of the parameters are repeated. This class also
        automatically assumes pair-reversibility and handles the setting of
        unlike pairs automatically i.e. A-B and B-A are set at the same time.

        Note that, unlike the :class:`pyPRISM.core.MatrixArray`, this
        container is not meant to be used for mathematics. The benefit of this
        is that, for each type, it can contain any arbitrary number, string, or
        Python object. 

        See the example below and the `pyPRISM Internals` section of the
        :ref:`tutorial` for more information.

    Example
    -------
    .. code-block:: python

        import pyPRISM

        PT = pyPRISM.PairTable(['A','B','C'],name='potential')

        # Set the 'A-A' pair
        PT['A','A']            = 'Lennard-Jones'

        # Set the 'B-A', 'A-B', 'B-B', 'B-C', and 'C-B' pairs
        PT['B',['A','B','C'] ] = 'Weeks-Chandler-Andersen'

        # Set the 'C-A', 'A-C', 'C-C' pairs
        PT['C',['A','C'] ]     = 'Exponential'

        for i,t,v in PT.iterpairs():
            print('{}) {} for pair {}-{} is {}'.format(i,VT.name,t[0],t[1],v))

        # The above loop prints the following:
        #   (0, 0)) potential for pair A-A is Lennard-Jones
        #   (0, 1)) potential for pair A-B is Weeks-Chandler-Andersen
        #   (0, 2)) potential for pair A-C is Exponential
        #   (1, 1)) potential for pair B-B is Weeks-Chandler-Andersen
        #   (1, 2)) potential for pair B-C is Weeks-Chandler-Andersen
        #   (2, 2)) potential for pair C-C is Exponential

        for i,t,v in PT.iterpairs(full=True):
            print('{}) {} for pair {}-{} is {}'.format(i,VT.name,t[0],t[1],v))

        # The above loop prints the following:
        #   (0, 0)) potential for pair A-A is Lennard-Jones
        #   (0, 1)) potential for pair A-B is Weeks-Chandler-Andersen
        #   (0, 2)) potential for pair A-C is Exponential
        #   (1, 0)) potential for pair B-A is Weeks-Chandler-Andersen
        #   (1, 1)) potential for pair B-B is Weeks-Chandler-Andersen
        #   (1, 2)) potential for pair B-C is Weeks-Chandler-Andersen
        #   (2, 0)) potential for pair C-A is Exponential
        #   (2, 1)) potential for pair C-B is Weeks-Chandler-Andersen
        #   (2, 2)) potential for pair C-C is Exponential

    
    
    '''
    def __init__(self,types,name,symmetric=True):
        r'''Constructor

        Arguments
        ----------
        types: list
            Lists of the types that will be used to key the PairTable. The length of this
            list should be equal to the rank of the PRISM problem to be solved, i.e. 
            len(types) == number of sites in system
            
        name: string
            The name of the PairTable. This is simply used as a convencience for identifying
            the table internally. 
        
        symmetric: bool
            If *True*, the table will automatically set both off-diagonal values during
            assignment e.g. PT['A','B'] = 5 will set 'A-B' and 'B-A'
        '''
        self.types = types
        self.symmetric = symmetric
        self.name = name
        self.values = {t1:{t2:None for t2 in types} for t1 in types}
    
    def __repr__(self):
        return '<PairTable: {}>'.format(self.name)
        
    def __iter__(self):
        for (i,t1),(j,t2) in product(enumerate(self.types),enumerate(self.types)):
            yield (i,j),(t1,t2),self.values[t1][t2]
            
    def __getitem__(self,index):
        t1,t2 = index
        return self.values[t1][t2]
    
    def __setitem__(self,index,value):
        types1,types2 = index
        for t1 in self.listify(types1):
            for t2 in self.listify(types2):
                
                # If we don't copy the value, later modifications to this element
                # can affect all other set items. While this could be used intentionally,
                # it negates a primary use case of setting a global "default" value across
                # the table and then only modifying specific elements afterwards
                value_copy = copy.deepcopy(value) 
                
                self.values[t1][t2] = value_copy
                if self.symmetric and t1!=t2:
                    self.values[t2][t1] = value_copy
            
    def check(self):
        '''Is everything in the table set?

        Raises
        ------
        ValueError if all values are not set
        
        '''
        for i,t,val in self.iterpairs():
            if val is None:
                raise ValueError('PairTable {} is not fully specified!'.format(self.name))
            
            
    def iterpairs(self,full=False,diagonal=True):
        '''Convenience function for looping over table pairs.
        
        Parameters
        ----------
        full: bool
            If *True*, all i,j pairs (upper and lower diagonal) will be looped over
            
        diagonal: bool
            If *True*, only the i==j (on-diagonal) pairs will be considered when looping
        
        '''
        
        if full:
            test = lambda i,j: True
        elif diagonal:
            test = lambda i,j: i<=j
        else:
            test = lambda i,j: i<j
            
        for (i,j),(t1,t2),val in self.__iter__():
            if test(i,j):
                yield (i,j),(t1,t2),(val)
                
    def setUnset(self,value):
        '''Set all values that have not been specified to a value

        Arguments
        ---------
        value: 
            Any valid python object (number, list, array, etc) can be passed in
            as a value for all unset fields. 
        
        '''
        for i,(t1,t2),v in self.iterpairs():
            if v is None:
                self[t1,t2] = value
                
    def exportToMatrixArray(self,space=Space.Real):
        '''Convenience function for converting a table of arrays to a MatrixArray

        .. warning::

            This only works if the PairTable contains numerical data that is
            all of the same shape that can be cast into a np.ndarray like
            object.
        
        '''
        lengths = []
        for i,t,val in self.iterpairs():
            if val is None:
                raise ValueError('Can\'t export not-fully specified Table {}'.format(self.name))
            lengths.append(len(val))
            
        if not len(set(lengths))<=1:
            raise ValueError('Arrays in Table are not all the same length. Aborting export.')
        
        length = lengths[0]
        rank = len(self.types)
        MA = MatrixArray(length=length,rank=rank,space=space,types=self.types)
        
        for i,(t1,t2),val in self.iterpairs():
            MA[t1,t2] = val
        return MA
    
    def apply(self,func,inplace=True):
        '''Apply a function to all elements in the table in place
        
        Parameters
        ----------
        func: any object with __call__ method
            function to be called on all table elements

        inplace: bool
            If *True*, apply modifications to self. Otherwise, create a new PairTable.
        
        '''
        if inplace:
            table = self
        else:
            table = PairTable(types=self.types,name=self.name,symmetric=self.symmetric)
            
        for i,(t1,t2),val in self.iterpairs():
            table[t1,t2] = func(val)
            
        return table
        

        



#!python
from __future__ import division,print_function
from pyPRISM.core.Space import Space

def pair_correlation(PRISM):
    r'''Calculate the Real-space *inter*-molecular pair correlation function 

    Parameters
    ----------
    PRISM: pyPRISM.core.PRISM
        A **solved** PRISM object.
    
    Returns
    -------
    pairCorr: pyPRISM.core.MatrixArray
        The full MatrixArray of pair correlation functions.
    

    **Mathematical Definition**

    .. math::
         
        g_{\alpha,\beta}(r) = h_{\alpha,\beta}(r) + 1.0
    
    **Variable Definitions**

        - :math:`g_{\alpha,\beta}(r)`
            Pair correlation function between site types :math:`\alpha` and
            :math:`\beta` at a distance :math:`r`

        - :math:`h_{\alpha,\beta}(r)`
            Total correlation function between site types :math:`\alpha` and
            :math:`\beta` at a distance :math:`r`

    **Description**

        The pair correlation function describes the spatial correlations
        between pairs of sites in Real-space. Also known as the *radial
        distribution function* (rdf), the :math:`g(r)` function is
        related to the underlying spatial probability distributions of a given
        system. In a PRISM calculation, :math:`g(r)` is strictly an
        *inter*-molecular quantity.

        After convergence of a PRISM object, the stored total correlation
        attribute function can simply be shifted to obtain the :math:`g(r)` 


    .. warning::

        Passing an unsolved PRISM object to this function will still produce
        output based on the default values of the attributes of the PRISM
        object.
    

    Example
    -------
    .. code-block:: python

        import pyPRISM

        sys = pyPRISM.System(['A','B'])
        
        # ** populate system variables **
        
        PRISM = sys.createPRISM()

        PRISM.solve()

        rdf = pyPRISM.calculate.pair_correlation(PRISM)

        rdf_AA = rdf['A','A']
        rdf_AB = rdf['A','B']
        rdf_BB = rdf['B','B']
    
    '''
    
    if PRISM.totalCorr.space == Space.Fourier:
        PRISM.sys.domain.MatrixArray_to_real(PRISM.totalCorr)
    
    PRISM.pairCorr = PRISM.totalCorr + 1.0
    
    return PRISM.pairCorr



#!python
from __future__ import division,print_function
from pyPRISM.closure.AtomicClosure import AtomicClosure
import numpy as np
import warnings 
class MeanSphericalApproximation(AtomicClosure):
    r'''Mean Spherical Approximation closure

    **Mathematial Definition**

        .. math:: c_{\alpha,\beta}(r) = -U_{\alpha,\beta}(r)


    **Variables Definitions**

        - :math:`c_{\alpha,\beta}(r)`
            Direct correlation function value at distance :math:`r` between
            sites :math:`\alpha` and :math:`\beta`.
    
        - :math:`U_{\alpha,\beta}(r)`
            Interaction potential value at distance :math:`r` between sites
            :math:`\alpha` and :math:`\beta`.
    

    **Description**
        
        The Mean Spherical Approximation (MSA) closure assumes an interaction
        potential that contains a hard-core interaction and a tail interaction.
        See Reference [1] for a derivation and discussion of this closure.
        
        The MSA does a good job of describing the properties of the square-well 
        fluid, and allows for the analytical solution of the PRISM/RISM 
        equations for some systems. The MSA closure reduces to the PercusYevick 
        closure if the tail is ignored.

    
    References
    ----------
    #. Hansen, J.P.; McDonald, I.R.; Theory of Simple Liquids; Chapter 4, Section 4; 
       4th Edition (2013), Elsevier [`link
       <https://www.sciencedirect.com/science/book/9780123870322>`__]

    Example
    -------
    .. code-block:: python

        import pyPRISM

        sys = pyPRISM.System(['A','B'])
        
        sys.closure['A','A'] = pyPRISM.closure.PercusYevick()
        sys.closure['A','B'] = pyPRISM.closure.PercusYevick()
        sys.closure['B','B'] = pyPRISM.closure.MeanSphericalApproximation()

        # ** finish populating system object **

        PRISM = sys.createPRISM()

        PRISM.solve()
    
    '''
    def __init__(self,apply_hard_core=False):
        '''Contstructor

        Parameters
        ----------
        apply_hard_core: bool
            If *True*, the total correlation function will be assumed to be -1
            inside the core (:math:`r_{i,j}<(d_i + d_j)/2.0`) and the closure
            will not be applied in this region.
        '''
        self.potential = None
        self.value = None
        self.sigma = None
        self.apply_hard_core = apply_hard_core

        if apply_hard_core == False:
            warnings.warn(
                    '''The MSA closure does not work for divergent potentials
                    when the hard core condition is not manually applied. This
                    will likely result in a cryptic crash of the simulation if
                    attempted. Using MSA(apply_hard_core=True) will avoid this
                    warning. This warning should be ignored if hard-core
                    interactions are not being used.''')

        
    def __repr__(self):
        return '<AtomicClosure: MeanSphericalApproximation>'
    
        
    def calculate(self,r,gamma):
        '''Calculate direct correlation function based on supplied :math:`\gamma`

        Arguments
        ---------
        r: np.ndarray
            array of real-space values associated with :math:`\gamma`

        gamma: np.ndarray
            array of :math:`\gamma` values used to calculate the direct
            correlation function
        
        '''
        
        assert self.potential is not None,'Potential for this closure is not set!'
        
        assert len(gamma) == len(self.potential),'Domain mismatch!'
        
        if self.apply_hard_core:
            # apply hard core condition 
            self.value = -1 - gamma

            # calculate closure outside hard core
            mask = r>self.sigma
            self.value[mask] = -self.potential[mask]
        else:
            self.value = -self.potential

        
        return self.value

class MSA(MeanSphericalApproximation):
    '''Alias of MeanSphericalApproximation'''
    pass



#!python
from pyPRISM.core.MatrixArray import MatrixArray
from pyPRISM.core.Space import Space
import numpy as np

class IdentityMatrixArray(MatrixArray):
    '''Specialization of MatrixArray which is initialized with Identity matrices

    See :class:`pyPRISM.core.MatrixArray` for details
    
    '''
    
    def __init__(self,length,rank,data=None,space=None,types=None):
        super(IdentityMatrixArray,self).__init__(length=length,rank=rank,space=space,types=types)

        self.data = np.zeros((length,rank,rank))
        for i in range(rank):
            self.data[:,i,i] = 1.0
        



#!python
from __future__ import division,print_function
from pyPRISM.potential.Potential import Potential
import numpy as np

class Exponential(Potential):
    r'''Exponential attractive interactions
    
    
    **Mathematical Definition**
    
    .. math::
    
        U_{\alpha,\beta}(r \geq \sigma_{\alpha,\beta}) = \epsilon_{\alpha,\beta} \exp\left(- \frac{r-\sigma_{\alpha,\beta}}{\alpha}\right)

    .. math::

        U_{\alpha,\beta}(r < \sigma_{\alpha,\beta}) = C^{high}

    
    **Variable Definitions**
    
    :math:`\alpha`
        Width of exponential attraction

    :math:`\sigma_{\alpha,\beta}`
        Contact distance of interactions between sites 
	:math:`\alpha` and :math:`\beta`.


    :math:`\epsilon_{\alpha,\beta}`
        Interaction strength between sites 
	:math:`\alpha` and :math:`\beta`.

    :math:`C^{high}`
        High value used to approximate an infinite potential due to overlap
    

    **Description**

    	This potential models an exponential-like attraction between sites with
        a specified site size and contact distance. For example, in Reference
        [1] this potential is used to model the attraction between a 
	nanoparticle and monomers of a polymer chain. 


    References
    ----------
    #. Hooper, J.B. and K.S. Schweizer, Theory of phase separation in polymer
       nanocomposites. Macromolecules, 2006. 39(15): p. 5133-5142.
       [`link <https://doi.org/10.1021/ma060577m>`__]

    

    Example
    -------
    .. code-block:: python

        import pyPRISM
	
        #Define a PRISM system and set the A-B interaction potential
	sys = pyPRISM.System(['A','B'],kT=1.0)
	sys.domain = pyPRISM.Domain(dr=0.1,length=1024)
        sys.potential['A','B'] = pyPRISM.potential.Exponential(epsilon=1.0,sigma=8.0,alpha=0.5,high_value=10**6)

    .. warning::

        If sigma is specified such that it does not fall on the solution grid
        of the :class:`~pyPRISM.core.Domain.Domain` object specified in
        :class:`~pyPRISM.core.System.System`, then the sigma will effectively
        be rounded. A warning should be emitted during the construction of a
        :class:`~pyPRISM.core.PRISM.PRISM` object if this occurs.
    
    '''
    def __init__(self,epsilon,alpha,sigma=None,high_value=1e6):
        r''' Constructor
        
        Arguments
        ---------
        epsilon: float
            Strength of attraction
            
        alpha: float
            Range of attraction

        sigma: float, *optional*
            Contact distance. If not specified, sigma will be calculated from 
            the diameters specified in the :class:`~pyPRISM.core.System.System`
            object.
            
        high_value: float, *optional*
            High value used to approximate an infinite potential due to overlap
        
        '''
        self.epsilon = epsilon
        self.sigma = sigma
        self.alpha = alpha
        self.high_value = high_value
        self.funk  = lambda r,sigma: - epsilon * np.exp(-(r-sigma)/(alpha))
    def __repr__(self):
        return '<Potential: Exponential>'
    
    def calculate(self,r):
        r'''Calculate the value of the potential

        Attributes
        ----------
        r: float np.ndarray
            Array of pair distances at which to calculate potential values
        '''
        assert (self.sigma is not None), 'Sigma must be set before evaluating potential!'

        magnitude = self.funk(r,self.sigma)
        magnitude = np.where(r>self.sigma,magnitude,self.high_value)
        return magnitude
        



#!python
r'''
Once the PRISM equation is solved, :math:`\hat{H}(k)`, :math:`\hat{\Omega}(k)`, and
:math:`\hat{C}(k)` are used to calculate various structural and thermodynamic
properties.  As listed below, pyPRISM provides functions that calculate
these properties from solved :class:`pyPRISM.core.PRISM` objects. 

If a desired calculation is not listed, please consider filing an `Issue
<https://github.com/usnistgov/pyPRISM/issues>`_ on GitHub, implementing the
calculation, and sharing with the community. See :ref:`contribute` for more
details.  

'''

from pyPRISM.calculate.solvation_potential import solvation_potential
from pyPRISM.calculate.pair_correlation    import pair_correlation
from pyPRISM.calculate.structure_factor    import structure_factor
from pyPRISM.calculate.second_virial       import second_virial
from pyPRISM.calculate.pmf                 import pmf
from pyPRISM.calculate.chi                 import chi
from pyPRISM.calculate.spinodal_condition  import spinodal_condition



#!python
from __future__ import division,print_function
from pyPRISM.closure.AtomicClosure import AtomicClosure
import numpy as np
class HyperNettedChain(AtomicClosure):
    r'''HyperNettedChain closure 

    **Mathematial Definition**

        .. math:: c_{\alpha,\beta}(r) = \exp\left(\gamma_{\alpha,\beta}(r)-U_{\alpha,\beta}(r)\right) - 1.0 -  \gamma_{\alpha,\beta}(r)

        .. math:: \gamma_{\alpha,\beta}(r) =  h_{\alpha,\beta}(r) - c_{\alpha,\beta}(r)

    
    **Variables Definitions**

        - :math:`h_{\alpha,\beta}(r)` 
            Total correlation function value at distance :math:`r` between
            sites :math:`\alpha` and :math:`\beta`.
    
        - :math:`c_{\alpha,\beta}(r)`
            Direct correlation function value at distance :math:`r` between
            sites :math:`\alpha` and :math:`\beta`.
    
        - :math:`U_{\alpha,\beta}(r)`
            Interaction potential value at distance :math:`r` between sites
            :math:`\alpha` and :math:`\beta`.
    
    **Description**

        The Hypernetted Chain Closure (HNC) is derived by expanding the
        direct correlation function, :math:`c(r)`, in powers of density shift
        from a reference state. See Reference [1] for a full derivation and
        discussion of this closure.
        
        The change of variables is necessary in order to use potentials with
        hard cores in the computational setting. Written in the standard form,
        this closure diverges with divergent potentials, which makes it
        impossible to numerically solve. 

        Compared to the PercusYevick closure, the HNC closure is a more
        accurate approximation of the full expression for the direct
        correlation function. Depsite this, it can produce inaccurate,
        long-range fluctuations that make it difficult to employ in
        phase-separating systems. The HNC closure performs well for systems
        where there is a disparity in site diameters and is typically used for
        the larger site. 
    
    References
    ----------
    #. Hansen, J.P.; McDonald, I.R.; Theory of Simple Liquids; Chapter 4, Section 4; 
       4th Edition (2013), Elsevier [`link
       <https://www.sciencedirect.com/science/book/9780123870322>`__]

    Example
    -------
    .. code-block:: python

        import pyPRISM

        sys = pyPRISM.System(['A','B'])
        
        sys.closure['A','A'] = pyPRISM.closure.PercusYevick()
        sys.closure['A','B'] = pyPRISM.closure.PercusYevick()
        sys.closure['B','B'] = pyPRISM.closure.HypernettedChain()

        # ** finish populating system object **

        PRISM = sys.createPRISM()

        PRISM.solve()
    
    '''
    def __init__(self,apply_hard_core=False):
        '''Contstructor

        Parameters
        ----------
        apply_hard_core: bool
            If *True*, the total correlation function will be assumed to be -1
            inside the core (:math:`r_{i,j}<(d_i + d_j)/2.0`) and the closure
            will not be applied in this region. Defaults to *True*.
        '''
        self.potential = None
        self.value = None
        self.sigma = None
        self.apply_hard_core = apply_hard_core
        
    def __repr__(self):
        return '<AtomicClosure: HyperNettedChain>'
    
    def calculate(self,gamma):
        '''Calculate direct correlation function based on supplied :math:`\gamma`

        Arguments
        ---------
        gamma: np.ndarray
            array of :math:`\gamma` values used to calculate the direct
            correlation function
        
        '''
        
        assert self.potential is not None,'Potential for this closure is not set!'
        
        assert len(gamma) == len(self.potential),'Domain mismatch!'
        
        
        return self.value

        
        
    def calculate(self,r,gamma):
        '''Calculate direct correlation function based on supplied :math:`\gamma`

        Arguments
        ---------
        r: np.ndarray
            array of real-space values associated with :math:`\gamma`

        gamma: np.ndarray
            array of :math:`\gamma` values used to calculate the direct
            correlation function
        
        '''
        
        assert self.potential is not None,'Potential for this closure is not set!'
        
        assert len(gamma) == len(self.potential),'Domain mismatch!'
        
        if self.apply_hard_core:
            assert self.sigma is not None, 'If apply_hard_core=True, sigma parameter must be set!'

            # apply hard core condition 
            self.value = -1 - gamma

            # calculate closure outside hard core
            mask = r>self.sigma
            self.value[mask] = np.exp(gamma[mask] - self.potential[mask]) - 1.0 - gamma[mask]
        else:
            self.value = np.exp(gamma - self.potential) - 1.0 - gamma

        
        return self.value
        
        
class HNC(HyperNettedChain):
    '''Alias of HyperNettedChain'''
    pass



#!python
from __future__ import division,print_function
from pyPRISM.omega.Gaussian import Gaussian
import unittest
import numpy as np

class Gaussian_TestCase(unittest.TestCase):
    def test_create(self):
        '''Can we create a Gaussian Omega?'''
        k = np.arange(0.75,3.5,0.05)
        sigma = 1.0
        length = N = 100
        
        E = np.exp(-k*k*sigma*sigma/6.0)
        N = length
        O1 = (1 - E*E - 2*E/N + (2*E**(N+1))/N)/((1-E)**2.0)
        
        G = Gaussian(sigma,length)
        O2 = G.calculate(k)
        np.testing.assert_array_almost_equal(O1,O2)
        
        



#!python
from __future__ import division,print_function
from pyPRISM.potential.Potential import Potential
import numpy as np
class HardCoreLennardJones(Potential):
    r'''12-6 Lennard-Jones potential with Hard Core

    .. warning:: 

        This potential uses a slightly different form than what is 
        implemented for the classic LJ potential. This means that the epsilon
        in the LJ and HCLJ potentials will not correspond to the same
        interaction strengths.

    **Mathematical Definition**
    
    .. math::
    
        U_{\alpha,\beta}(r>\sigma_{\alpha,\beta}) = \epsilon_{\alpha,\beta}\left[\left(\frac{\sigma_{\alpha,\beta}}{r}\right)^{12} - 2 \left(\frac{\sigma_{\alpha,\beta}}{r}\right)^{6}\right]
    
    .. math::
    
        U_{\alpha,\beta}(r\leq\sigma_{\alpha,\beta}) = C^{high}
    
    
    **Variable Definitions**
    
    :math:`\epsilon_{\alpha,\beta}`
        Strength of interaction (attraction or repulsion) between sites 
	:math:`\alpha` and :math:`\beta`.

    :math:`\sigma_{\alpha,\beta}`
        Length scale of interaction between sites 
	:math:`\alpha` and :math:`\beta`.

    :math:`r`
        Distance between sites. 

    :math:`C^{high}`
        High value used to approximate an infinite potential due to overlap
    
   
    **Description**

    	Unlike the classic LJ potential, the HCLJ potential has an infinitely
    	hard core and can handle negative and positive epsilons, corresponding
    	to attractive and repulsive interactions.

    References
    ----------
    #. Yethiraj, A. and K.S. Schweizer, INTEGRAL-EQUATION THEORY OF POLYMER
       BLENDS - NUMERICAL INVESTIGATION OF MOLECULAR CLOSURE APPROXIMATIONS.
       Journal of Chemical Physics, 1993. 98(11): p. 9080-9093.
       [`link <https://doi.org/10.1063/1.464466>`__]
    
    
    Example
    -------
    .. code-block:: python

        import pyPRISM
	
        #Define a PRISM system and set the A-B interaction potential
	sys = pyPRISM.System(['A','B'],kT=1.0)
	sys.domain = pyPRISM.Domain(dr=0.1,length=1024)
        sys.potential['A','B'] = pyPRISM.potential.HardCoreLennardJones(epsilon=1.0,sigma=1.0,high_value=10**6)

    .. warning::

        If sigma is specified such that it does not fall on the solution grid
        of the :class:`~pyPRISM.core.Domain.Domain` object specified in
        :class:`~pyPRISM.core.System.System`, then the sigma will effectively
        be rounded. A warning should be emitted during the construction of a
        :class:`~pyPRISM.core.PRISM.PRISM` object if this occurs.
    
    '''
    def __init__(self,epsilon,sigma=None,high_value=1e6):
        r''' Constructor
        
        Arguments
        ---------
        epsilon: float
            Depth of attractive well
            
        sigma: float, *optional*
            Contact distance. If not specified, sigma will be calculated from 
            the diameters specified in the :class:`~pyPRISM.core.System.System`
            object.
            
        high_value: float, *optional*
            High value used to approximate an infinite potential due to overlap
        
        '''
        self.epsilon = epsilon 
        self.sigma = sigma
        self.high_value = high_value
        self.funk  = lambda r,sigma: epsilon * ((sigma/r)**(12.0) - 2.0*(sigma/r)**(6.0))
        
    def __repr__(self):
        return '<Potential: HardCoreLennardJones>'
        
    def calculate(self,r):
        r'''Calculate value of potential

        Attributes
        ----------
        r: float np.ndarray
            Array of pair distances at which to calculate potential values
        '''
        assert (self.sigma is not None), 'Sigma must be set before evaluating potential!'

        magnitude = self.funk(r,self.sigma)
        magnitude[r<=self.sigma] = self.high_value
                
        return magnitude
        



#!python
from __future__ import division,print_function
from pyPRISM.omega.FromArray import FromArray
import unittest
import numpy as np
import os

class FromArray_TestCase(unittest.TestCase):
    def test_create(self):
        '''Can we create a FromArray Omega?'''
        k = np.arange(0.75,3.5,0.05)
        omega = np.ones_like(k)
        
        
        FF = FromArray(k=k,omega=omega).calculate(k)
        
    def test_error_check(self):
        '''Does FromFile correctly check the domain?'''
        fname = 'FromFileTestCase.dat'

        k1 = np.arange(0.75,3.5,0.05)
        k2 = np.arange(0.75,3.5,0.01)
        k3 = np.arange(0.75,4.0,0.05)
        omega = np.ones_like(k1)

        
        with self.assertRaises(AssertionError):
            FF = FromArray(k=k1,omega=omega).calculate(k2)

        with self.assertRaises(AssertionError):
            FF = FromArray(k=k1,omega=omega).calculate(k3)
        
        
        
if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(FromArray_TestCase)
    unittest.TextTestRunner(verbosity=2).run(suite)



#!python

from __future__ import division,print_function
import numpy as np

class MolecularClosure:
    '''Baseclass for all *molecular* closures
    
    .. note::

        Currently, this class doesn't do anything besides group all of the
        *molecular* closures under a single inheritance heirarchy. 
    '''
        



#!python
from __future__ import division,print_function
from pyPRISM.closure import HyperNettedChain
import unittest
import numpy as np

class HyperNettedChain_TestCase(unittest.TestCase):
    def test_create(self):
        '''Can we create and use  a HyperNettedChain closure?'''
        r = np.arange(0.75,3.5,0.05)
        U = np.zeros_like(r)
        U[r<1.0] = 1e6
        
        gamma = np.zeros_like(r)
        C1 = np.exp(gamma-U) - 1.0 - gamma
        
        
        HNC = HyperNettedChain(apply_hard_core=False)
        HNC.potential = U
        C2 = HNC.calculate(r,gamma)
        np.testing.assert_array_almost_equal(C1,C2)

    def test_create_hard_core(self):
        '''Can we create and use a hard-core HyperNettedChain closure?'''
        r = np.arange(0.75,3.5,0.05)
        U = np.zeros_like(r)
        U[r<1.0] = 1e6
        sigma = 1.0
        
        gamma = np.zeros_like(r)
        C1 = -1 - gamma
        mask = r>sigma
        C1[mask] = np.exp(gamma[mask]-U[mask]) - 1.0 - gamma[mask]
        
        
        HNC = HyperNettedChain(apply_hard_core=False)
        HNC.potential = U
        HNC.sigma = sigma
        C2 = HNC.calculate(r,gamma)
        np.testing.assert_array_almost_equal(C1,C2)
        
        
if __name__ == '__main__':
    import unittest 
    suite = unittest.TestLoader().loadTestsFromTestCase(HyperNettedChain_TestCase)
    unittest.TextTestRunner(verbosity=2).run(suite)



#!python
from __future__ import division,print_function
from pyPRISM.core.Space import Space

def structure_factor(PRISM,normalize=True):
    r'''Calculate the structure factor from a PRISM object
    
        
    Parameters
    ----------
    PRISM: pyPRISM.core.PRISM
        A **solved** PRISM object.

    normalize: bool
        normalize the structure factor by the site density
    
    Returns
    -------
    structureFactor: pyPRISM.core.MatrixArray
        The full MatrixArray of structure factors

        
    **Mathematical Definition**


    .. math::

        \hat{s}_{\alpha,\beta}(k) = \rho^{site}_{\alpha,\beta} \hat{\omega}_{\alpha,\beta}(k) + \rho^{pair}_{\alpha,\beta} \hat{h}_{\alpha,\beta}(k)

    .. math::

        \hat{s}_{\alpha,\beta}^{norm}(k) = \hat{s}_{\alpha,\beta}(k) / \rho^{site}_{\alpha,\beta}  

    
    **Variable Definitions**

        - :math:`\hat{\omega}_{\alpha,\beta}(k)`
            Intra-molecular correlation function between sites :math:`\alpha`
            and :math:`\beta` at a wavenumber :math:`k`

        - :math:`\hat{h}_{\alpha,\beta}(k)`
            Total correlation function between sites :math:`\alpha` and
            :math:`\beta` at a wavenumber :math:`k`

        - :math:`\rho^{site}_{\alpha,\beta}`, :math:`\rho^{pair}_{\alpha,\beta}`
            Sitewise and pairwise densities for sites :math:`\alpha` and
            :math:`\beta`. See :class:`pyPRISM.core.Density` for details. 

    **Description**

        The structure factor (:math:`\hat{s}_{\alpha,\beta}(k)`) is a 
        Fourier-space representation of the structural correlations between
        sites :math:`\alpha` and :math:`\beta`. The :math:`\hat{s}_{\alpha,\beta}(k)`
        can be related to the real-space pair correlation function through a 
        Fourier transform. In the PRISM formalism, the 
        :math:`\hat{s}_{\alpha,\beta}(k)` can be calculated as the sum of the 
        Fourier-space intra-molecular and total correlation functions, as 
        shown above. 


    .. warning::

        Passing an unsolved PRISM object to this function will still produce
        output based on the default values of the attributes of the PRISM
        object.
    
    
    References
    ----------
    #. Chandler, D., Introduction to Modern Statistical Mechanics, 
       Oxford U. Press, New York, 1987 [`link
       <https://books.google.com/books/about/Introduction_to_Modern_Statistical_Mecha.html?id=3taTh5D-CDsC>`__]

    #. Schweizer, Curro, Integral equation theory of the structure and
       thermodynamics of polymer blends, J. Chem. Phys., 1989 91 (8) 5059 [`link
       <https://doi.org/10.1063/1.457598>`__]


    Example
    -------
    .. code-block:: python

        import pyPRISM

        sys = pyPRISM.System(['A','B'])

        # ** populate system variables **
        
        PRISM = sys.createPRISM()

        PRISM.solve()

        sk = pyPRISM.calculate.structure_factor(PRISM)

        sk_BB = sk['B','B']
    
    '''
    
    if PRISM.totalCorr.space == Space.Real:
        PRISM.sys.domain.MatrixArray_to_fourier(PRISM.totalCorr)
        
    if PRISM.omega.space == Space.Real:
        PRISM.sys.domain.MatrixArray_to_fourier(PRISM.omega)
    
    structureFactor = (PRISM.totalCorr*PRISM.sys.density.pair + PRISM.omega)/PRISM.sys.density.site
    # structureFactor = PRISM.totalCorr + PRISM.omega
    
    return structureFactor



#!python
from __future__ import division,print_function
import os
import unittest
import numpy as np
from pyPRISM.trajectory.Debyer import Debyer
from pyPRISM import Domain



class Debyer_TestCase(unittest.TestCase):
    def setup_rod(self,Nf,N):
        positions = []
        for j in range(Nf):
            dummy = []
            for i in range(N):
                dummy.append([0.0,0.0,i])
            positions.append(dummy)
        positions = np.array(positions)
        molecules = np.zeros(N,dtype=int)
        box = np.repeat([[1000.,1000.,1000.]],Nf,axis=0)
        return positions,molecules,box
    def test_self_omega(self):
        '''Can we calculate a self-omega?'''

        domain = Domain(dk = 0.1, length = 1024)
        debyer = Debyer(domain=domain,nthreads=2)

        Nf = 1
        N = 50
        positions,molecules,box = self.setup_rod(Nf,N)

        selfHist = True
        omega = debyer.calculate(positions,positions,
                                 molecules,molecules,
                                 box,
                                 selfHist)

        base_path = os.path.split(__file__)[0]
        file_path = os.path.join(base_path,'data','Omega-Test-1Rod.dat')
        k, omega_control = np.loadtxt(file_path)
        np.testing.assert_array_almost_equal(omega,omega_control,decimal=3)
    def test_non_self_omega(self):
        '''Can we calculate a non-self omega?'''

        domain = Domain(dk = 0.1, length = 1024)
        debyer = Debyer(domain=domain,nthreads=2)

        Nf = 1
        N = 50
        positions,molecules,box = self.setup_rod(Nf,N)

        mask1 = (np.arange(N)<(N//2))
        mask2 = (np.arange(N)>=(N//2))

        positions1 = positions[:,mask1,:]
        positions2 = positions[:,mask2,:]
        molecules1 = molecules[mask1]
        molecules2 = molecules[mask2]

        selfHist = False
        omega = debyer.calculate(positions1,positions2,
                                 molecules1,molecules2,
                                 box,
                                 selfHist)

        base_path = os.path.split(__file__)[0]
        file_path = os.path.join(base_path,'data','Omega-Test-2Rod.dat')
        k, omega_control = np.loadtxt(file_path)
        np.testing.assert_array_almost_equal(omega,omega_control)

        

        
if __name__ == '__main__':
    import unittest 
    suite = unittest.TestLoader().loadTestsFromTestCase(Debyer_TestCase)
    unittest.TextTestRunner(verbosity=2).run(suite)



#!python
from __future__ import division,print_function
from pyPRISM.core.ValueTable import ValueTable
from pyPRISM.core.PairTable import PairTable
from pyPRISM.core.MatrixArray import MatrixArray
from pyPRISM.core.Space import Space
import numpy as np

class Density(object):
    r'''Container for pair and site densities

    **Mathematical Definition**

    .. math::
        
        \rho^{pair}_{\alpha,\beta} = \rho_{\alpha} \rho_{\beta} 

    .. math::
        
        \rho^{site}_{\alpha,\beta} = 
            \begin{cases}
                \rho_{\alpha}                & \text{if } i = j \\
                \rho_{\alpha} + \rho_{\beta} & \text{if } i \neq j
            \end{cases}

    .. math::
        
        \rho^{total} = \sum_{\alpha} \rho^{site}_{\alpha,\alpha}


    **Variable Definitions**

        :math:`\rho_{\alpha}`
            Number density of site :math:`\alpha`

        :math:`\rho^{pair}_{\alpha,\beta}`
            Pair number density of pair :math:`\alpha,\beta`

        :math:`\rho^{site}_{\alpha,\beta}`
            Site number density of pair :math:`\alpha,\beta`

        :math:`\rho^{total}`
            Total site number density 
        

    **Description**
    
        This class describes the makeup of the system in terms of both total
        site and pair densities. The container provides a simple interface for
        getting and setting (via square brackets [ ]) site densities and also
        takes care of calculating the total site and total pair number
        densities. The total site and pair number densities can be accessed as
        MatrixArrays (:class:`pyPRISM.core.MatrixArray`) attributes. 
    

    Example
    -------
    .. code-block:: python

        import pyPRISM

        rho = pyPRISM.Density(['A','B','C'])

        rho['A'] = 0.25
        rho['B'] = 0.35
        rho['C'] = 0.15

        rho.pair['A','B'] #pair density rho_AB = 0.25 * 0.35
        rho.site['A','B'] #site density rho_AB = 0.25 + 0.35
        rho.site['B','B'] #site density rho_BB = 0.35
        rho.total         #total density rho   = 0.25 + 0.35 + 0.15
    
    '''
    def __init__(self,types):
        r'''Constructor 

        Arguments
        ---------
        types: list 
            List of types of sites
        
        Attributes
        ----------
        density: :class:`pyPRISM.core.ValueTable`
            Table of site number density values

        total: float
            Total number density 

        site: :class:`pyPRISM.core.MatrixArray`
            Site density for each pair.

        pair: :class:`pyPRISM.core.MatrixArray`
            Pair site density for each pair.
        '''
        self.types = types 

        self.density = ValueTable(types=types,name='density')
        self.total = 0.

        self.pair = MatrixArray(length=1,rank=len(types),types=types,space=Space.NonSpatial)
        self.site = MatrixArray(length=1,rank=len(types),types=types,space=Space.NonSpatial)

    def check(self):
        '''Are all densities set?

        Raises
        ------
        *ValueError* if densities are not all set. 
        
        '''
        self.density.check()

    def __repr__(self):
        return '<Density total:{:3.2f}>'.format(self.total)

    def __getitem__(self,key):
        return self.density[key]

    def __setitem__(self,types1,value):
        for t1 in self.density.listify(types1):
            rho1 = value
            self.density[t1] = rho1

            self.total = 0.
            for t2 in self.types:
                # If rho2 isn't set yet, we can't set the
                # site or pair densities
                rho2 = self.density[t2]
                if rho2 is None:
                    continue

                self.total += rho2

                # The values must be set as lists in order for them to
                # be compatible with the MatrixArray type
                self.pair[t1,t2] = [rho1*rho2]
                if t1 == t2:
                    self.site[t1,t2] = [rho1]
                else:
                    self.site[t1,t2] = [rho1 + rho2]
